/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.kex.bwsin.konfig.kb.KonfigurationsObjekt;
import de.bsvrz.kex.bwsin.konfig.kb.SINSTS;

/**
 * F&uuml;hrt diverse Tests auf den ASB-Stationierungsangaben der neuen
 * Konfiguration durch und gibt die Ergebnisse als Datei aus.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: AsbTests.java 10460 2008-07-10 17:28:00Z gieseler $
 * 
 */

public class AsbTests {

	static class AsbProblem {

		public static final AsbProblem KEINE_ASBSTATIONIERUNG = new AsbProblem(
				"keine ASB-Stationierung");

		public static final AsbProblem KEINE_ASBRICHTUNG = new AsbProblem(
				"keine ASB-Stationierungsrichtung");

		public static final AsbProblem OFFSET_FEHLER = new AsbProblem(
				"Offset-Fehler");

		public static final AsbProblem LAENGEN_FEHLER = new AsbProblem(
				"Lngen-Fehler");

		private final String problemtext;

		/**
		 * @param string
		 */
		public AsbProblem(String string) {
			problemtext = string;
		}

		/**
		 * {@inheritDoc}
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return problemtext;
		}
	}

	class AsbProblemObjekt {

		KonfigurationsObjekt objekt;
		List<String> probleme = new ArrayList<String>();

		/**
		 */
		public AsbProblemObjekt(KonfigurationsObjekt objekt, String problem) {
			this.objekt = objekt;
			probleme.add(problem);
		}

		public void addProblem(String problem) {
			probleme.add(problem);
		}

		/**
		 * {@inheritDoc}
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			String text = objekt.getPid() + "(" + objekt.getName() + "):";
			for (String p : probleme) {
				text += " " + p;
			}

			return text;
		}
	}

	Map<String, AsbProblemObjekt> aMap = new HashMap<String, AsbProblemObjekt>();

	private void addProblem(KonfigurationsObjekt ko, String problem) {
		AsbProblemObjekt asbobjekt = aMap.get(ko.getPid());

		if (asbobjekt == null) {
			aMap.put(ko.getPid(), new AsbProblemObjekt(ko, problem));
		} else {
			asbobjekt.addProblem(problem);
		}
	}

	/**
	 * @param dateiname
	 * @param objekte
	 */
	public void erzeugeInfo(final String dateiname,
			final Collection<KonfigurationsObjekt> objekte) {
		PrintStream ps;
		AsbProblemObjekt aobjekt;
		List<String> asbliste = new ArrayList<String>();
		try {
			ps = new PrintStream(new File(dateiname));
		} catch (FileNotFoundException e) {
			System.err.println("Fehler beim Anlegen der Datei: " + dateiname);
			e.printStackTrace();
			return;
		}

		for (KonfigurationsObjekt o : objekte) {
			if (o instanceof SINSTS) {
				SINSTS sts = (SINSTS) o;
				if (sts.getAsbStationierungen().size() > 1) {
					asbliste.add(o.getPid());
				}
				if (sts.getAsbStationierungen().size() == 0) {
					addProblem(o, AsbProblem.KEINE_ASBSTATIONIERUNG.toString());
				}
				int asblaenge = 0;
				for (AsbStationierung asb : sts.getAsbStationierungen()) {
					if (asb.getVerkehrsRichtung() == Verkehrsrichtung.RICHTUNG_UNBEKANNT) {
						addProblem(o, AsbProblem.KEINE_ASBRICHTUNG.toString());
					}
					if (asb.getOffset() > sts.getLaenge()) {
						addProblem(o, AsbProblem.OFFSET_FEHLER.toString());
					}
					asblaenge += asb.getStationierungEnde()
							- asb.getStationierungAnfang();
				}
				if (asblaenge > 0) {
					double abw = sts.getLaenge() / asblaenge;
					if ((abw > 1.07) || (abw < 0.93)) {
						addProblem(o, AsbProblem.LAENGEN_FEHLER.toString()
								+ "(" + abw + ")");
					}
				} else {
					if (sts.getLaenge() > 1) {
						addProblem(o, AsbProblem.LAENGEN_FEHLER.toString()
								+ "(" + sts.getLaenge() + "bei '0' ASB-Lnge)");
					}
				}

			}
		}

		ps
				.println("=====================================================================================");
		ps
				.println("Strassenteilsegmente mit mehr als einer ASB-Stationierung:");
		ps
				.println("=====================================================================================");
		for (String s : asbliste) {
			ps.println(s);
		}

		ps.println();

		ps
				.println("=====================================================================================");
		ps.println("Strassenteilsegmente mit ASB-Stationierungsproblemen");
		ps
				.println("=====================================================================================");

		for (AsbProblemObjekt objekt : aMap.values()) {
			ps.println(objekt);
		}
		ps.close();
	}
}
