/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraBetriebsKilometer;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repr&auml;sentiert die Betriebskilometer-Features als Map mit dem
 * Schl&uuml;ssel Abschnitt oder Ast. Fr jeden Abschnitt oder Ast wird eine
 * nach der Stationierung geordnete Liste der zugeh&ouml;rigen Betriebskilometer
 * angelegt.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: BetriebsKilometer.java 10402 2008-07-10 09:07:00Z gieseler $
 * 
 */
public class BetriebsKilometer {

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private final Debug debug = Debug.getLogger();

	/**
	 * Map der Betriebskilometer-Punkte.
	 */
	private final Map<String, HashMap<Integer, BetriebsKilometerPunkt>> bkMap = new HashMap<String, HashMap<Integer, BetriebsKilometerPunkt>>();

	/**
	 * F&uuml;gt einen Betriebskilometer-Punkt zur Map hinzu.
	 * 
	 * @param bk
	 *            Betriebskilometer-Punkt
	 * @return true, wenn neu hinzugef&uuml;gt, sonst false
	 */
	private boolean addPunkt(final BetriebsKilometerPunkt bk) {
		HashMap<Integer, BetriebsKilometerPunkt> pmap = bkMap.get(bk.getAoa());
		boolean neu = false;

		if (bkMap.get(bk.getAoa()) == null) {
			pmap = new HashMap<Integer, BetriebsKilometerPunkt>();
			bkMap.put(bk.getAoa(), pmap);
		}

		BetriebsKilometerPunkt tfp = pmap
				.get(new Integer(bk.getStationierung()));
		if (tfp == null) {
			pmap.put(new Integer(bk.getStationierung()), bk);
			tfp = bk;
			neu = true;
		} else {
			debug.error("Doppelter Betriebskilometer AoA " + //$NON-NLS-1$
					bk.getAoa() + " Stationierung " + //$NON-NLS-1$
					bk.getStationierung());
		}

		return neu;
	}

	/**
	 * Liefert eine Liste von Betriebskilometerpunkten, die auf einem Abschnitt
	 * oder Ast liegen.
	 * 
	 * @param aoa
	 *            Abschnitt oder Ast
	 * @return Liste von Betriebskilometerpunkten auf diesem AoA, aufsteigend
	 *         nach Stationierung sortiert
	 */
	public List<BetriebsKilometerPunkt> getAoaPunkte(final String aoa) {
		List<BetriebsKilometerPunkt> rlist = new ArrayList<BetriebsKilometerPunkt>();
		if (bkMap.get(aoa) != null) {
			rlist.addAll(bkMap.get(aoa).values());

			Collections.sort(rlist, new Comparator<BetriebsKilometerPunkt>() {
				public int compare(final BetriebsKilometerPunkt o1,
						final BetriebsKilometerPunkt o2) {
					return new Integer(o1.getStationierung())
							.compareTo(new Integer(o2.getStationierung()));
				}
			});
		}

		return rlist;
	}

	/**
	 * Liefert eine Liste von Betriebskilometerpunkten, die auf einem Abschnitt
	 * oder Ast in einem bestimmten Stationierungsbereichbereich liegen.
	 * 
	 * @param aoa
	 *            Abschnitt oder Ast
	 * @param stationierunganfang
	 *            Anfangsstationierung
	 * @param stationierungende
	 *            Endstationierung
	 * @return Liste von Betriebskilometerpunkten auf diesem AoA im angefragten
	 *         Stationierungsbereich
	 */
	public List<BetriebsKilometerPunkt> getAoaPunkte(final String aoa,
			int stationierunganfang, int stationierungende) {
		List<BetriebsKilometerPunkt> aoalist = getAoaPunkte(aoa);
		List<BetriebsKilometerPunkt> rlist = new ArrayList<BetriebsKilometerPunkt>();

		for (BetriebsKilometerPunkt bkp : aoalist) {
			if (bkp.getStationierung() >= stationierunganfang
					&& bkp.getStationierung() <= stationierungende) {
				rlist.add(bkp);
			}
		}

		return rlist;
	}

	/**
	 * Initialisierungsfunktion f&uuml;r die Betriebskilometer.
	 * 
	 * @param okstradb
	 *            die Okstra-Datenbasis
	 * @return Anzahl der erzeugten Betriebskilometer-Punkte
	 */
	public int initialisiere(final Okstra okstradb) {
		int added = 0;
		Map<String, Feature> fmap;
		Feature feature;

		debug.info("Betriebskilometer werden initialisiert."); //$NON-NLS-1$

		try {
			fmap = okstradb.getFeatureMap(OkstraFeatures.BETRIEBSKILOMETER
					.getFeatureOkstraName());

			for (Map.Entry<String, Feature> e : fmap.entrySet()) {
				feature = e.getValue();

				String aoaID;
				try {
					aoaID = Okstra.getInstance().getSpecialAttribute2(
							OkstraFeatures.BETRIEBSKILOMETER, feature.getID(),
							OkstraBetriebsKilometer.AOA);
				} catch (ImportException e2) {
					debug.error(e2.getMessage());
					continue;
				}

				if ((aoaID == null) || aoaID.length() == 0) {
					debug
							.error("Das Betriebskilometer-Feature " + feature.getID() + //$NON-NLS-1$
									" besitzt keine AoA-Stationierungsangabe"); //$NON-NLS-1$
					continue;
				}

				Feature bkaoa = Okstra.getFeatureReferenz(aoaID);
				if (bkaoa == null) {
					debug.error("Das referenzierte Feature " + aoaID + //$NON-NLS-1$
							" existiert nicht"); //$NON-NLS-1$
					continue;
				}

				try {
					if (addPunkt(new BetriebsKilometerPunkt(bkaoa.getID(),
							feature))) {
						added++;
					}
				} catch (ImportException e1) {
					debug.error(e1.getMessage());
					continue;
				}

			}

			debug.fine("Erzeugt: " + added + " Betriebskilometer-Feature(s)"); //$NON-NLS-1$ //$NON-NLS-2$

		} catch (final IllegalArgumentException e1) {
			debug.error(e1.getMessage());
		} catch (ImportException e) {
			debug.error(e.getMessage());
		}

		return added;
	}

}
