/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Netzrepr&auml;sentation des zu importierenden Netzes auf der Basis der
 * Verbindungspunkte. &Uuml;ber diese Klasse k&ouml;nnen die
 * Stra&szlig;enabschnitte und die inneren Stra&szlig;ensegmente gebildet
 * werden.
 * 
 * Das Netz wird zun&auml;chst mit den Stra&szlig;enelementen der OKSTRA-
 * Datenbasis initialisiert. Die Stra&szlig;enelemente werden dabei entsprechend
 * ihrer Verkehrsrichtung den Verbindungspunkten zugeordnet. Stra&szlig;en-
 * elemente, die in beiden Verkehrsrichtungen definiert sind, werden gedoppelt.
 * 
 * Die gebildeten Stra&szlig;enabschnitte werden ebenfalls den
 * Verbindungspunkten zugewiesen. &Uuml;ber den Stra&szlig;enabschnitte
 * k&ouml;nnen dann die inneren Stra&szlig;ensegmente durch Bildung der Routen
 * gefunden werden.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: ImportNetz.java 10402 2008-07-10 09:07:00Z gieseler $
 * 
 */

public class ImportNetz extends TreeMap<String, Verbindungspunkt> {
	/**
	 * serialVersionUID.
	 */
	private static final long serialVersionUID = -4706045995500520386L;

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private final Debug debug = Debug.getLogger();

	/**
	 * HashSet der importierten Stra&szlig;enelemente.
	 */
	private final ArrayList<StrassenElement> selemente = new ArrayList<StrassenElement>();

	/**
	 * Die verbotenen Fahrbeziehungen.
	 */
	VerboteneFahrbeziehungen verboteneFahrbeziehungen = null;

	/**
	 * gefundene Routen.
	 */
	private ArrayList<Route> routen;

	/**
	 * Konstruktor aus OKSTRA-Datenbasis. Es wird ein Netz aus den
	 * Verbindungspunkten der OKSTRA-Datenbasis und den Stra&szlig;enelementen
	 * erzeugt. Die Zuordnung der Stra&szlig;enelemente erfolgt entsprechend
	 * ihrer Verkehrsrichtung, d.h. bei Stra&szlig;enelementen mit der
	 * Verkehrsrichtung 'GEGEN_STATIONIERUNGSRICHTUNG' werden die Ver-
	 * bindungspunkte getauscht. Stra&szlig;enelemente mit Verkehrsrichtung
	 * 'BEIDE_RICHTUNGEN' werden ge- doppelt und ein Element f&uuml;r jede
	 * Richtung in das Netz integriert.
	 * 
	 * @param okstradb
	 *            {@link Okstra}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public ImportNetz(Okstra okstradb) throws ImportException {

		debug.info("Das Importnetz wird erzeugt");

		// Verbindungspunkte einlesen
		Map<String, Feature> ovpunkte;
		ovpunkte = okstradb.getFeatureMap(OkstraFeatures.VERBINDUNGSPUNKTE
				.getFeatureOkstraName());

		// bilde HashMap Verbindungspunkte
		for (Map.Entry<String, Feature> e : ovpunkte.entrySet()) {
			Verbindungspunkt v = new Verbindungspunkt(e.getValue());
			this.put(e.getKey(), v);
		}

		importiereStrassenElemente(okstradb);

		// zuordnen Strassenelemente zu V-Punkten
		Iterator<StrassenElement> it = selemente.iterator();

		while (it.hasNext()) {
			StrassenElement se = it.next();
			Verbindungspunkt vp = this.get(se.getStartVP());
			if (vp == null) {
				debug.error("Strassenelement '" + se.getID() + //$NON-NLS-1$
						"' beginnt bei nicht vorhandenem Verbindungspunkt"); //$NON-NLS-1$
				continue;
			}

			vp.addVon(se);

			vp = this.get(se.getEndeVP());

			if (vp == null) {
				debug.error("Strassenelement '" + se.getID() + //$NON-NLS-1$
						"' endet bei nicht vorhandenem Verbindungspunkt"); //$NON-NLS-1$
			} else {
				vp.addNach(se);
			}
		}

		verboteneFahrbeziehungen = new VerboteneFahrbeziehungen();
		verboteneFahrbeziehungen.initialisiere(okstradb);
		debug.info("Das Importnetz wurde erzeugt");

	}

	/**
	 * Erzeugt aus einem Stra&szlig;enelement f&uuml;r jede Location, auf die es
	 * zuf&auml;hrt, eine eigene Instanz.
	 * 
	 * @param element
	 *            {@link StrassenElement}
	 * @return Liste von {@link StrassenElement}
	 */
	private List<StrassenElement> bildeStrassenElementeAusLocations(
			StrassenElement element) {
		List<StrassenElement> liste = new ArrayList<StrassenElement>();

		if (element.istASSFeature()) {
			try {
				List<Location> locations = element.getFeatureNachLocationen();
				if (locations.size() == 0) {
					debug.error("Location-Inkonsistenz: " + element.getID());
				} else if (locations.size() == 1) {
					element.setNachLocation(locations.get(0));
					liste.add(element);
				} else {
					// bilde Strassenelemente mit jeweils einer Location
					for (int i = 0; i < locations.size(); i++) {
						StrassenElement sel;
						if (i == 0) {
							sel = element;
						} else {
							sel = element.clone();
						}

						sel.setNachLocation(locations.get(i));
						liste.add(sel);
					}
				}

			} catch (ImportException e) {
				debug.error(e.getMessage());
			}
		} else {
			liste.add(element);
		}

		return liste;
	}

	/**
	 * L&ouml;scht die 'benutzt'-Flags aller Verbindungspunkte.
	 */
	private void cleanupUsed() {
		for (Verbindungspunkt vp : this.values()) {
			vp.setUsed(false);
		}
	}

	/**
	 * Erzeugt einen Strassenabschnitt beginnend bei einem Stra&szlig;enelement.
	 * Mehrere Stra&szlig;enelemente werden unter folgenden Bedingungen zu einem
	 * Stra&szlig;enabschnitt zusammengefa&szlig;t: 1. Die Verbindungspunkte an
	 * den Enden des Stra&szlig;enelementes haben genau ein eingehendes
	 * Stra&szlig;enelement (endet am Verbindungspunkt) und genau ein abgehendes
	 * Stra&szlig;enelemente (beginnt bei Verbindungs- punkt). 2. Die
	 * TMC-Richtungsangaben der zu verbindenden Stra&szlig;enelemente sind
	 * gleich.
	 * 
	 * Der Algoritmus arbeitet wie folgt: Vom Start-Stra&szlig;enelement aus
	 * wird zuerst vorw&auml;rts und dann r&uuml;ckw&auml;rts &uuml;ber die
	 * Verbindungspunkte nach Stra&szlig;enelementen gesucht, die o.g. Kriterien
	 * entsprechen.
	 * 
	 * @param startelement
	 *            Strassenelement
	 * @return neuer Strassenabschnitt
	 */
	private StrassenAbschnitt erzeugeAbschnitt(StrassenElement startelement) {
		// neuer Abschnitt fr dieses Element
		StrassenAbschnitt abschnitt = new StrassenAbschnitt(startelement);

		abschnitt.setRichtung(startelement.getRichtung());

		startelement.zuAbschnitt(abschnitt);

		// nur ASS verbinden
		if (!startelement.istASS()) {
			return abschnitt;
		}

		// gehe vorwrts
		StrassenElement strelement = startelement;
		for (;;) {
			// Verbindungspunkt, der das Element als ankommend enthlt
			Verbindungspunkt vpnach = this.get(strelement.getEndeVP());
			if (vpnach == null) {
				debug.error("Strassenelement '" + strelement.getID() + //$NON-NLS-1$
						"' endet bei nicht vorhandenem Verbindungspunkt"); //$NON-NLS-1$
				break;
			}

			// suche aktuelles Element in der Liste der eingehenden Elemente des
			// Zielpunktes
			int found = 0;
			for (StrassenElement se : vpnach.getElementeAnkommend()) {
				if (strelement == se) {
					found++;
					break;
				}
			}

			if (found == 0) {
				debug.error("inkonsistentes Netz: Strassenelement '" + //$NON-NLS-1$
						strelement.getID() + "' , nicht ankommend bei VP: " + //$NON-NLS-1$
						vpnach.getFeature().getID());
				return null;
			}

			StrassenElement naechsteselement = vpnach
					.findeNachfolger(strelement);
			if (naechsteselement == null) {
				break;
			}

			// Test, ob die Straenelemente verbunden werden knnen
			if (!abschnitt.kannVerbundenWerden(naechsteselement)) {
				break;
			}

			debug.finest("Strassenelemente '" + //$NON-NLS-1$
					abschnitt.getLast().getID() + "' und '" + //$NON-NLS-1$
					naechsteselement.getID() + "' werden verbunden (Ende)"); //$NON-NLS-1$

			abschnitt.anfuegenEnde(naechsteselement);
			naechsteselement.zuAbschnitt(abschnitt);
			strelement = naechsteselement;
		}

		// gehe rckwrts
		strelement = startelement;
		for (;;) {
			// Start-Verbindungspunkt
			Verbindungspunkt vpvon = this.get(strelement.getStartVP());
			if (vpvon == null) {
				debug.error("Strassenelement '" + strelement.getID() + //$NON-NLS-1$
						"' beginnt bei nicht vorhandenem Verbindungspunkt"); //$NON-NLS-1$
				break;
			}

			StrassenElement naechsteselement = vpvon
					.findeVorgaenger(strelement);
			if (naechsteselement == null) {
				break;
			}

			// Test, ob die Straenelemente verbunden werden knnen
			if (!abschnitt.kannVerbundenWerden(naechsteselement)) {
				break;
			}

			if (!naechsteselement.kannBenutztWerden()) {
				debug.error("inkonsistentes Netz: Strassenelement '" + //$NON-NLS-1$
						strelement.getID() + "' , VP: " + //$NON-NLS-1$
						vpvon.getID() + ", prev: " + naechsteselement.getID()); //$NON-NLS-1$
				break;
			}

			debug.finest("Strassenelement '" + //$NON-NLS-1$
					naechsteselement.getID() + "' und '" + //$NON-NLS-1$
					abschnitt.getFirst().getID()
					+ "' werden verbunden (Anfang)"); //$NON-NLS-1$

			abschnitt.anfuegenAnfang(naechsteselement);
			naechsteselement.zuAbschnitt(abschnitt);
			strelement = naechsteselement;
		}

		return abschnitt;
	}

	/**
	 * Bildet Stra&szlig;enabschnitte aus zusammengehrenden
	 * Stra&szlig;enelementen ber den Verbindungspunkten.
	 * 
	 * @return Liste von StrassenAbschnitten
	 */
	public List<StrassenAbschnitt> erzeugeStrassenabschnitte() {
		List<StrassenAbschnitt> retlist = new ArrayList<StrassenAbschnitt>();

		int anzahlelemente = selemente.size();

		int proc = 0;

		Iterator<StrassenElement> li = selemente.iterator();
		while (li.hasNext()) {
			// finde ein unbenutztes Strassenelement
			StrassenElement se = li.next();
			while (se.kannBenutztWerden()) {
				debug.fine("bearbeite: " + se.getID() + //$NON-NLS-1$
						"(" + ++proc + ")"); //$NON-NLS-1$ //$NON-NLS-2$
				StrassenAbschnitt sa = erzeugeAbschnitt(se);

				if (sa != null) {
					retlist.add(sa);
					zuweisenStrassenabschnitt(sa);
				}
			}
		}

		debug.info("erzeugt: " + retlist.size() + " Straenabschnitte aus " + //$NON-NLS-1$//$NON-NLS-2$
				anzahlelemente + " Strassenelementen"); //$NON-NLS-1$

		// Testausgabe
		// System.out.println("Anzahl Abschnitte: " + retlist.size());
		// for (StrassenAbschnitt ab : retlist) {
		// if (ab.size() > 1) {
		// System.out.println("Abschnitt " + ab.getRichtung() + " mit "
		// + ab.size() + " Strassenelemente(n)");
		// String abd = new String();
		// for (StrassenElement se : ab) {
		// abd += "[" + se.getFeature().getID() + " (VVP "
		// + se.get_startVP() + ", NVP " + se.get_endeVP()
		// + ")] ";
		// }
		// System.out.println(abd);
		// }
		//
		// if (ab.size() == 0) {
		// System.out.println("Abschnitt mit " + ab.size()
		// + " Strassenelemente(n)");
		// }
		// }

		return retlist;
	}

	/**
	 * Sucht ein an einem Verbindungspunkt ankommendes ASS.
	 * 
	 * @param vpunktid
	 *            Feature-ID des Verbindungspunktes
	 * @return Stra&szlig;enabschnitt oder null, wenn kein ankommendes ASS
	 *         existiert
	 */
	public StrassenAbschnitt findeAssAnkommend(final String vpunktid) {
		final Verbindungspunkt start = this.get(vpunktid);
		for (final StrassenAbschnitt sa : start.getStrassenAbschnitteNach()) {
			if (sa.istASS()) {
				return sa;
			}
		}

		return null;
	}

	/**
	 * Findet weitere Stra&szlig;enelemente zu einer Route ausgehend von einem
	 * Verbindungspunkt.
	 * 
	 * @param vpunktid
	 *            ID des Start-Verbindungspunktes
	 * @param route
	 *            Route
	 */
	private void findeInnereStrassenelemente(String vpunktid, Route route) {
		boolean routenende = true;

		Verbindungspunkt start = this.get(vpunktid);
		if (start.isUsed()) {
			// Schleife erkannt -> stop
			debug.finest("Schleife erkannt an " + start.getID()); //$NON-NLS-1$
			routen.remove(route);
			return;
		}

		debug.finest("beginne routing ab " + vpunktid); //$NON-NLS-1$

		start.setUsed(true);

		for (StrassenAbschnitt sa : start.getStrassenAbschnitteVon()) {
			Route froute = new Route(route);

			if (verboteneFahrbeziehungen.istVerboten(froute, sa)) {
				debug.finest("verbotene Fahrbeziehung an " + vpunktid
						+ " nach " + sa); //$NON-NLS-1$
				continue;
			}

			// Test auf Rckweg auf gleichem Straenelement
			if (route.getLast().getLast().getID().equals(sa.getFirst().getID())) {
				debug.finest("Wende an " + vpunktid);
				continue;
			}

			routenende = false;

			// Test auf Schleife
			if (this.get(sa.endeVP).isUsed()) {
				debug.finest("interne Schleife erkannt an "
						+ this.get(sa.endeVP).getID());
				continue;
			}

			// stop bei ASS
			if (sa.istASS()) {
				froute.add(sa);
				routen.add(froute);
				continue;
			}

			debug.finest("an " + vpunktid + " wird " + sa + " angefgt"); //$NON-NLS-1$
			froute.add(sa);
			findeInnereStrassenelemente(sa.letzterVP(), froute);
		}

		if (routenende) {
			debug.finest("keine weiteren Abschnitte"); //$NON-NLS-1$
			routen.add(route);
		}

		start.setUsed(false);
	}

	/**
	 * Findet die inneren Stra&szlig;ensegmente beginnend bei einem
	 * Stra&szlig;enabschnitt. Es werden alle Verbindungen ausgehend von dem
	 * Stra&szlig;enabschnitt gesucht, bis entweder ein &auml;u&szlig;eres
	 * Stra&szlig;enelement erreicht wird oder kein weiterer
	 * Stra&szlig;enabschnitt mehr folgt. Das Ergebnis ist eine Liste von
	 * Routen, wobei jede Route durch eine Liste von Stra&szlig;enabschnitten
	 * gebildet wird. Die berechneten Routen werden auf enthaltene verbotene
	 * Fahrbeziehungen getestet und bei Bedarf korrigiert. Die Routen werden am
	 * Ende geordnet. Die Ordnung wird durch den ber die Methode
	 * {@link ImportNetz} erzeugten Comparator bestimmt.
	 * 
	 * @param startabschnitt
	 *            Stra&szlig;enabschnitt, mit dem das Routing beginnt
	 * @return Liste von Routen
	 */
	public synchronized ArrayList<Route> findeInnereStrassenSegmente(
			StrassenAbschnitt startabschnitt) {
		routen = new ArrayList<Route>();
		cleanupUsed();
		Route route = new Route();

		debug.finest("beginne routing ab " + startabschnitt); //$NON-NLS-1$
		route.add(startabschnitt);

		this.get(startabschnitt.startVP).setUsed(true);
		findeInnereStrassenelemente(startabschnitt.letzterVP(), route);

		// berechnete Routen auf verbotene Fahrbeziehungen testen
		ArrayList<Route> ergebnisrouten = new ArrayList<Route>();

		for (Route troute : routen) {
			Route korrroute = verboteneFahrbeziehungen.bildeErlaubtenAnfang(
					troute, this);

			if (korrroute != troute) {
				debug
						.fine("Die Route '"
								+ troute
								+ "' wurde aufgrund verbotener Fahrbeziehungen korrigiert");
				if (korrroute.size() > 0) {
					debug.fine("Neue Route '" + korrroute + "'");

					// TEST
					if (ergebnisrouten.contains(korrroute)) {
						debug.info("doppelte Route " + korrroute
								+ " wird verworfen");
					} else {
						ergebnisrouten.add(korrroute);
					}
				} else {
					debug.fine("Die Route wurde verworfen");
				}
			} else {
				ergebnisrouten.add(troute);
			}
		}

		// berechnete Routen ordnen
		Collections.sort(ergebnisrouten, getComparator());
		return ergebnisrouten;
	}

	/**
	 * Erzeugt einen Comparator zum Ordnen der erzeugten Routen. Die Routen
	 * werden nach ihrer Berechnung in der von diesem Comparator bestimmten
	 * Reihenfolge abgearbeitet. Gefordert ist eine Ordnung entsprechende der
	 * 'verkehrstechnischen Bedeutung'. Dies wird hier als L&auml;nge in dem
	 * Sinne interpretiert, dass die 'verkehrstechnische Bedeutung' mit
	 * zunehmender L&auml;nge abnimmt.
	 * 
	 * @return {@link Comparator} zum Vergleich zweier Routen
	 */
	private Comparator<Route> getComparator() {
		return new Comparator<Route>() {
			public int compare(final Route route1, final Route route2) {
				double laenge1 = route1.getLaenge();
				double laenge2 = route2.getLaenge();
				return new Double(laenge1).compareTo(new Double(laenge2));
			}
		};
	}

	/**
	 * Bestimmt die Nachfolger eines Abschnittes.
	 * 
	 * @param abschnitt
	 *            Der Stra&szlig;enabschnitt, dessen Nachfolger bestimmt werden
	 *            sollen
	 * @return die Liste aller Nachfolger
	 */
	public List<StrassenAbschnitt> getNachfolger(StrassenAbschnitt abschnitt) {
		return this.get(abschnitt.endeVP).getStrassenAbschnitteVon();
	}

	/**
	 * Bestimmt einen noch nicht zugeordneten Vorg&auml;ngerabschnitt zu einem
	 * Abschnitt.
	 * 
	 * @param strassenAbschnitt
	 *            Abschnitt, zu dem der Vorg&auml;nger bestimmt werden soll
	 * @return nicht zugeordneter Vorg&auml;ngerabschnitt oder null, falls kein
	 *         Vorg&auml;nger gefunden werden kann
	 */
	public StrassenAbschnitt getVorgaenger(StrassenAbschnitt strassenAbschnitt) {
		Verbindungspunkt vp = this.get(strassenAbschnitt.startVP);

		for (StrassenAbschnitt sa : vp.getStrassenAbschnitteNach()) {
			if (!sa.istZugeordnet()) {
				return sa;
			}
		}

		return null;
	}

	/**
	 * Importiert alle Stra&szlig;enelemente der OKSTRA-Datenbasis.
	 * 
	 * @param okstradb
	 *            OKSTRA-Datenbasis
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	private void importiereStrassenElemente(final Okstra okstradb)
			throws ImportException {
		// bilde HashSet Strassenelemente
		Map<String, Feature> oselemente = okstradb
				.getFeatureMap(OkstraFeatures.STRASSENELEMENT.getFeatureOkstraName());
		for (Feature f : oselemente.values()) {
			StrassenElement se;

			se = new StrassenElement(f);

			// ignoriere nicht importierbare Strassenelemente
			if (se.kannImportiertWerden()) {
				Verkehrsrichtung vrichtung = se.bestimmeRichtung();

				if (vrichtung == Verkehrsrichtung.RICHTUNG_BEIDE_RICHTUNGEN) {
					StrassenElement se2 = se.erzeugeGegenrichtung();

					List<StrassenElement> seliste = bildeStrassenElementeAusLocations(se2);
					// interner Hash-Test
					for (StrassenElement tse : seliste) {
						if (selemente.contains(tse)) {
							debug.error("Interner Hash-Fehler"); //$NON-NLS-1$
						}
					}

					selemente.addAll(seliste);
				}

				if (se.getRichtung() == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
					se.vpUmkehren();
				}

				List<StrassenElement> seliste = bildeStrassenElementeAusLocations(se);
				// interner Hash-Test
				for (StrassenElement tse : seliste) {
					if (selemente.contains(tse)) {
						debug.error("Interner Hash-Fehler"); //$NON-NLS-1$
					}
				}

				selemente.addAll(seliste);
			} else {
				debug.warning("Das Strassenelement '" + f.getID() + //$NON-NLS-1$
						"' kann nicht importiert werden."); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Bestimmt, ob ein Abschnitt ein g&uuml;ltiger Anfang eines Stra&szlig;en-
	 * segmentes aus einer verbotenen Fahrbeziehung ist.
	 * 
	 * @param abschnitt
	 *            {@link StrassenAbschnitt}
	 * @return true, wenn der Abschnitt ein g&uuml;ltiger Anfang eines
	 *         Stra&szlig;en- segmentes aus einer verbotenen Fahrbeziehung ist,
	 *         sonst false
	 */
	public boolean istNeuerAnfangAusVerbotenerFahrbeziehung(
			StrassenAbschnitt abschnitt) {
		return verboteneFahrbeziehungen
				.istNeuerAnfangAusVerbotenerFahrbeziehung(abschnitt);
	}

	/**
	 * Weist dem Netz einen Stra&szlig;enabschnitt zu.
	 * 
	 * @param abschnitt
	 *            Stra&szlig;enabschnitt
	 */
	private void zuweisenStrassenabschnitt(final StrassenAbschnitt abschnitt) {
		this.get(abschnitt.ersterVP()).addStrassenAbschnittVon(abschnitt);
		if (this.get(abschnitt.letzterVP()) == null) {
			debug.error("der letzte Verbindungspunkt des Abschnittes '"
					+ abschnitt + "' kann nicht bestimmt werden");
		} else {
			this.get(abschnitt.letzterVP()).addStrassenAbschnittNach(abschnitt);
		}
	}
}
