/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.kex.bwsin.konfig.aktuell.AufteilungspunktAktuelleKonfiguration;
import de.bsvrz.kex.bwsin.konfig.aktuell.AufteilungspunkteAktuelleKonfiguration;
import de.bsvrz.kex.bwsin.konfig.aktuell.SINAktuelleKonfiguration;
import de.bsvrz.kex.bwsin.konfig.kb.AbstraktStrassenSegment;
import de.bsvrz.kex.bwsin.konfig.kb.DummyISS;
import de.bsvrz.kex.bwsin.konfig.kb.KonfigurationsObjekt;
import de.bsvrz.kex.bwsin.konfig.kb.SINASS;
import de.bsvrz.kex.bwsin.konfig.kb.SINISS;
import de.bsvrz.kex.bwsin.konfig.kb.SINKonfigurationsBereich;
import de.bsvrz.kex.bwsin.konfig.kb.SINSTS;
import de.bsvrz.kex.bwsin.konfig.kb.SINStrasse;
import de.bsvrz.kex.bwsin.konfig.kb.SINStrassenKnoten;
import de.bsvrz.kex.bwsin.konfig.kb.SINStrassenSegment;
import de.bsvrz.kex.bwsin.konfig.lcl.LCL;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.vew.SINProperties;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modul KonfigurationsDaten. Das Modul f&uuml;hrt die zentralen Operationen des
 * Netzimportes aus.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: KonfigurationsDaten.java 10471 2008-07-11 11:21:53Z gieseler $
 * 
 */
public class KonfigurationsDaten {

	/**
	 * Logger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * Die gebildeten Stra&szlig;enabschnitte.
	 */
	List<StrassenAbschnitt> strassenAbschnitte = new ArrayList<StrassenAbschnitt>();

	/**
	 * Die erzeugten Konfigurationsobjekte.
	 */
	Map<String, KonfigurationsObjekt> konfigurationsObjekte = new TreeMap<String, KonfigurationsObjekt>();

	/**
	 * Das Importnetz.
	 */
	private ImportNetz importNetz;

	/**
	 * Die Betriebskilometerangaben aus OKSTRA.
	 */
	private BetriebsKilometer betriebsKilometer = null;

	/**
	 * Die Menge der ASS zugeordneten Stra&szlig;enabschnitte.
	 */
	private final HashSet<StrassenAbschnitt> assMenge = new HashSet<StrassenAbschnitt>();

	/**
	 * Die aktuelle Konfiguration.
	 */
	private final SINAktuelleKonfiguration aktuelleKonfiguration;

	/**
	 * Die Segment-Factory.
	 */
	private SegmentFactory segmentFactory = null;

	/**
	 * Datenverteiler-Verbindung.
	 */
	private final ClientDavConnection dav;

	/**
	 * OKSTRA-Datenbasis.
	 */
	private final Okstra okstradb;

	/**
	 * Die Aufeilungspunkte der aktuellen Konfiguration.
	 */
	AufteilungspunkteAktuelleKonfiguration apaktuell = null;

	/**
	 * Konstruktor.
	 * 
	 * @param konfiguration
	 *            aktuelle Konfiguration
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param okstradb
	 *            OKSTRA-Datenbasis
	 */
	public KonfigurationsDaten(final SINAktuelleKonfiguration konfiguration,
			final ClientDavConnection dav, final Okstra okstradb) {
		super();
		aktuelleKonfiguration = konfiguration;
		this.dav = dav;
		this.okstradb = okstradb;
		segmentFactory = new SegmentFactory(aktuelleKonfiguration);
	}

	/**
	 * Aufteilung der Stra&szlig;enabschnitte.
	 * <p>
	 * Die Stra&szlig;enabschnitte werden zun&auml;chst an den
	 * Aufteilungspunkten der OKSTRA-Datenbasis geteilt. Anschlie&szlig;end
	 * werden sie an den durch Koordinaten bestimmten Aufteilungspunkten der
	 * aktuellen Konfiguration geteilt.
	 * 
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public void abschnitteAufteilen() throws ImportException {
		TeilungspunkteOkstra apokstra = new TeilungspunkteOkstra();
		apokstra.initialisiere(okstradb);
		for (StrassenAbschnitt abschnitt : strassenAbschnitte) {
			abschnitt.erzeugeTeilAbschnitte(okstradb, apokstra);
		}

		List<AufteilungspunktAktuelleKonfiguration> punkte = apaktuell
				.getAufteilungspunkteKoordinaten();

		if (punkte.size() > 0) {
			DEBUG
					.info("Die Straenabschnitte werden an den Aufteilungpunkten der aktuellen Konfiguration aufgeteilt.");
			int index = 0;
			for (StrassenAbschnitt abschnitt : strassenAbschnitte) {
				DEBUG.finer("bearbeite " + abschnitt + " (" + ++index + " von "
						+ strassenAbschnitte.size() + ")");
				List<AufteilungspunktAktuelleKonfiguration> abschnittpunkte = abschnitt
						.aufteilenKoordinaten(punkte);
				punkte.removeAll(abschnittpunkte);
			}
		} else {
			DEBUG
					.info("Keine Aufteilungpunkte der aktuellen Konfiguration mit Koordinaten.");
		}

		// nicht benutzte Punkte sichern
		aktuelleKonfiguration
				.setNichtVerarbeiteteAufteilungspunkteKoordinaten(punkte);
	}

	/**
	 * F&uuml;gt die Teilsegmente eines Stra&szlig;ensegmentes zur Liste der
	 * Konfigurationsobjekte hinzu.
	 * 
	 * @param segment
	 *            Stra&szlig;ensegment
	 */
	private void addiereTeilsegmente(final SINStrassenSegment segment) {
		for (KonfigurationsObjekt ko : segment.getTeilSegmente()) {
			konfigurationsObjekte.put(ko.getPid(), ko);
		}
	}

	/**
	 * F&uuml;gt ein neu erstelltes Konfigurationsobjekt zur Tabelle der
	 * Konfigurationsobjekte hinzu. Es wird gepr&uuml;ft, ob das Objekt bereits
	 * existiert.
	 * 
	 * @param objekt
	 *            Kondigurationsobjekt
	 */
	void addKonfigurationsObjekt(final KonfigurationsObjekt objekt) {
		if (konfigurationsObjekte.containsKey(objekt.getPid())) {
			throw new RuntimeException(
					"Versuch ein existierendes Objekt erneut hinzuzufuegen");
		}

		konfigurationsObjekte.put(objekt.getPid(), objekt);
	}

	/**
	 * Bestimmt die Knotenlocation zu einem Stra&szlig;ensegment.
	 * 
	 * @param vonlc
	 *            Location, von der sich das Segment wegbewegt.
	 * @param nachlc
	 *            Location, auf die sich das Segment zubewegt.
	 * @return {@link Location}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	private Location bestimmeKnotenLocation(Location vonlc, Location nachlc)
			throws ImportException {
		if (vonlc != Location.LOCATION_UNBEKANNT) {
			return vonlc;
		}

		return nachlc.bestimmeVorgaenger();
	}

	/**
	 * Bildet die &auml;u&szlig;eren Stra&szlig;ensegmente.
	 * 
	 * @return Anzahl der gebildeten Segmente
	 */
	private int bildeAeussereStrassenSegmente() {
		int anzahl = 0;

		for (StrassenAbschnitt sa : strassenAbschnitte) {
			try {
				if (sa.istASS()) {
					Location loc = sa.getLocationNach();
					Location vonlc = loc.bestimmeVorgaenger();

					String pid = SegmentFactory.bildePIDASS(vonlc, loc);
					pid = segmentFactory.bildePIDStrassenSegment(pid, sa
							.getPolygon());

					String namenszusatz = ""; //$NON-NLS-1$

					// fr Testzwecke Namenszusatz setzen
					if (SINProperties.getInstance().istAktiv(
							SINProperties.PROP_KEY_XMLDEBUG)) {
						namenszusatz += sa.kurzInfo();
					}

					String name = SegmentFactory.bildeSegmentNamen(pid,
							namenszusatz);
					DEBUG.fine("erzeuge ASS: " + pid); //$NON-NLS-1$
					SINASS ass = new SINASS(pid, name);
					ass.addAbschnitt(sa);
					ass.setLocationVon(vonlc);
					ass.setLocationNach(loc);

					sa.addASS(ass);
					sa.setZugeordnet();

					assMenge.add(sa);

					addKonfigurationsObjekt(ass);

					anzahl++;
				}
			} catch (ImportException ie) {
				DEBUG.error(ie.getMessage());
			}
		}

		return anzahl;
	}

	/**
	 * Bildet ein inneres Stra&szlig;ensegment aus einer Route.
	 * 
	 * @param route
	 *            Liste von Stra&szlig;enabschnitten
	 * @param extraname
	 *            Namenserweiterung
	 * @param dummy
	 *            Flag, ob es sich um ein Dummy-ISS (eingef&uuml;gt zwischen 2
	 *            ASS) handelt
	 * @return inneres Stra&szlig;ensegment
	 */
	private SINISS bildeInneresStrassenSegment(Route route, String extraname,
			boolean dummy) {
		SINISS iss = null;

		try {
			Location vonlc = route.bestimmeLocationVon();
			Location nachlc = route.bestimmeLocationNach(importNetz);

			checkLocations(vonlc, nachlc);

			// Knotenlocation bestimmen
			Location knotenlocation = bestimmeKnotenLocation(vonlc, nachlc);

			// fr Testzwecke Namenszusatz setzen
			String namenszussatz = " " + extraname; //$NON-NLS-1$
			if (SINProperties.getInstance().istAktiv(
					SINProperties.PROP_KEY_XMLDEBUG)) {
				namenszussatz += " ("; //$NON-NLS-1$
			}

			// bilde die Liste der Abschnitte, die das ISS bilden durch
			// entfernen der begrenzenden ASS
			Route issroute = new Route();
			for (StrassenAbschnitt sa : route) {
				if (!sa.istASS()) {
					issroute.add(sa);

					// fr Testzwecke Namenszusatz
					if (SINProperties.getInstance().istAktiv(
							SINProperties.PROP_KEY_XMLDEBUG)) {
						namenszussatz += sa.kurzInfo();
					}
				}
			}

			String pid = SegmentFactory.bildePIDISS(vonlc, nachlc);
			pid = segmentFactory.bildePIDStrassenSegment(pid, issroute
					.getPolygon());
			String name = SINISS.bildeNamen(pid, knotenlocation);

			DEBUG.fine("erzeuge ISS: " + pid); //$NON-NLS-1$
			iss = new SINISS(pid, name);

			iss.setAbschnitte(issroute);

			// fr Testzwecke Namenszusatz
			if (SINProperties.getInstance().istAktiv(
					SINProperties.PROP_KEY_XMLDEBUG)) {
				namenszussatz += ")"; //$NON-NLS-1$
				iss.setName(name + namenszussatz);
			}

			// setze ein STS mit zwei Punktkoordinaten (gleicher Punkt) fr
			// Dummy-ISS
			if (dummy) {
				iss = new DummyISS(iss, route, segmentFactory);
			}

			iss.setLocationVon(vonlc);
			iss.setLocationNach(nachlc);

			iss.setKnotenLocation(knotenlocation);

			if (route.get(0).istASS()) {
				iss.setVonAss(route.get(0).getASSnach(vonlc));
			}

			if (nachlc != Location.LOCATION_UNBEKANNT) {
				if (route.getLast().istASS()) {
					iss.setNachAss(route.getLast().getASSnach(nachlc));
				} else {
					StrassenAbschnitt ende = route
							.bestimmeEndabschnitt(importNetz);
					if (ende != null) {
						iss.setNachAss(ende.getASSnach(nachlc));
					}
				}
			}
		} catch (ImportException e) {
			DEBUG.error("Das ISS kann nicht erzeugt werden (" + //$NON-NLS-1$
					e.getMessage() + ") Route: " + route); //$NON-NLS-1$
		}

		return iss;
	}

	/**
	 * Bildet innere Stra&szlig;ensegmente auf der Basis von Routen.
	 * 
	 * @param routen
	 *            Liste von Routen
	 * @return Anzahl der gebildeten inneren Teilsegmente
	 */
	private int bildeInnereStrassenSegmente(ArrayList<Route> routen) {
		int anzahl = 0;

		for (Route route : routen) {
			SINISS iss = null;
			if (route.size() == 1) {
				DEBUG.finer("Netzende: " + route.get(0).toString()); //$NON-NLS-1$
				continue;
			}

			if (route.get(0).istASS() && (route.size() == 2)
					&& route.getLast().istASS()) {
				DEBUG
						.warning("2 zusammenhngende ASS (" + route.get(0).kurzInfo() + //$NON-NLS-1$
								", " + route.get(1).kurzInfo() + ")"); //$NON-NLS-1$//$NON-NLS-2$
				// fr Testzwecke Dummy-ISS extra kennzeichnen
				iss = bildeInneresStrassenSegment(route, "Dummy zw. 2 ASS ("
						+ route.get(0).getLast().getID() + ","
						+ route.getLast().getFirst().getID() + ")", true);
				if (iss != null) {
					addKonfigurationsObjekt(iss);
					anzahl++;
				}
				continue;
			}

			iss = bildeInneresStrassenSegment(route, "", false); //$NON-NLS-1$
			if (iss != null) {
				addKonfigurationsObjekt(iss);
				anzahl++;
				for (StrassenAbschnitt sal : route) {
					sal.setZugeordnet();
				}
			}
		}

		return anzahl;
	}

	/**
	 * Bildet die Stra&szlig;en.
	 */
	public void bildeStrassen() {
		HashSet<String> locationMenge = new HashSet<String>();
		HashMap<String, KonfigurationsObjekt> strassen = new HashMap<String, KonfigurationsObjekt>();
		LCL lcl = LCL.getInstance();

		// sammle alle Locations, auf die sich ein Strassensegment
		// zubewegt
		for (KonfigurationsObjekt ko : konfigurationsObjekte.values()) {
			if ((ko instanceof SINASS) || (ko instanceof SINISS)) {
				String loc = ((SINStrassenSegment) ko).getLocationNach()
						.getLocationCode();
				if (loc.equals(Location.LOCATION_CODE_UNBEKANNT)) {
					continue;
				}

				String linienlocation = lcl.getLocationCodeWert(loc,
						LCL.LINEAR_REFERENZ);
				if (linienlocation == null || linienlocation.length() == 0) {
					DEBUG
							.error("Die Linienlocation des Straensegmentes " + ko.getPid() + //$NON-NLS-1$
									" kann nicht bestimmt werden."); //$NON-NLS-1$
					continue;
				}

				locationMenge.add(linienlocation);
			}
		}

		Iterator<String> it = locationMenge.iterator();

		while (it.hasNext()) {
			String linienlocation = it.next();
			if (strassen.get(linienlocation) != null) {
				continue;
			}

			// erzeuge Strasse
			SINStrasse strasse;
			try {
				strasse = SINStrasse.erzeugeStrasse(linienlocation);
			} catch (ImportException e) {
				DEBUG
						.error("Erzeugung Strasse fr Linienlocation " + linienlocation + " : " + e.getMessage()); //$NON-NLS-1$//$NON-NLS-2$
				continue;
			}
			strassen.put(linienlocation, strasse);

			addKonfigurationsObjekt(strasse);
		}

		// Strassen zu ASS zuordnen
		for (KonfigurationsObjekt ko : konfigurationsObjekte.values()) {
			if ((ko instanceof SINASS) || (ko instanceof SINISS)) {
				SINStrassenSegment seg = (SINStrassenSegment) ko;

				String loc = seg.getLocationNach().getLocationCode();

				if (loc.equals(Location.LOCATION_CODE_UNBEKANNT)) {
					continue;
				}

				String linienlocation = lcl.getLocationCodeWert(loc,
						LCL.LINEAR_REFERENZ);
				if (linienlocation == null || linienlocation.length() == 0) {
					DEBUG
							.error("Die Linienlocation des Straensegmentes " + ko.getPid() + //$NON-NLS-1$
									" kann nicht bestimmt werden."); //$NON-NLS-1$
					continue;
				}

				if (strassen.get(linienlocation) == null) {
					DEBUG
							.error("Das Strassenobjekt zum Locationcode " + linienlocation + //$NON-NLS-1$
									" kann nicht gefunden werden"); //$NON-NLS-1$
					continue;
				}

				seg.setZuStrasse(strassen.get(linienlocation));
			}
		}
	}

	/**
	 * Bildet die Stra&szlig;enabschnitte auf der Basis der OKSTRA-Daten.
	 * 
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public void bildeStrassenAbschnitte() throws ImportException {
		importNetz = new ImportNetz(okstradb);
		strassenAbschnitte = importNetz.erzeugeStrassenabschnitte();
		betriebsKilometer = new BetriebsKilometer();
		betriebsKilometer.initialisiere(okstradb);
	}

	/**
	 * Bildet die Str&szlig;enknoten.
	 * 
	 */
	public void bildeStrassenKnoten() {
		HashSet<String> locationSet = new HashSet<String>();
		HashMap<String, KonfigurationsObjekt> knotenmap = new HashMap<String, KonfigurationsObjekt>();

		// sammle alle Locations der Strassensegmente
		for (KonfigurationsObjekt ko : konfigurationsObjekte.values()) {
			if ((ko instanceof SINASS) || (ko instanceof SINISS)) {
				locationSet.add(((SINStrassenSegment) ko).getLocationNach()
						.getLocationCode());
				locationSet.add(((SINStrassenSegment) ko).getLocationVon()
						.getLocationCode());
			}
		}

		Iterator<String> it = locationSet.iterator();

		while (it.hasNext()) {
			String location = it.next();
			if (location.equals(Location.LOCATION_UNBEKANNT.getLocationCode())) {
				continue;
			}

			if (knotenmap.get(location) != null) {
				// bereits vorhanden
				continue;
			}

			// erzeuge Strassenknoten
			SINStrassenKnoten strassenknoten = SINStrassenKnoten
					.erzeugeStrassenKnoten(dav, location);

			for (String knotenlocation : strassenknoten.getLocationen()) {
				knotenmap.put(knotenlocation, strassenknoten);
			}

			addKonfigurationsObjekt(strassenknoten);
		}

		for (KonfigurationsObjekt ko : konfigurationsObjekte.values()) {
			if (ko instanceof SINASS) {
				SINASS ass = (SINASS) ko;
				Location von = ass.getLocationVon();
				ass.setVonKnoten(knotenmap.get(von.getLocationCode()));

				Location nach = ass.getLocationNach();
				ass.setNachKnoten(knotenmap.get(nach.getLocationCode()));
			}

			if (ko instanceof SINISS) {
				SINISS iss = (SINISS) ko;
				Location knoten = iss.getKnotenLocation();
				if (knoten == null) {
					DEBUG.error("ISS " + iss.getPid()
							+ " hat keine Knotenlocation.");
					continue;
				}

				SINStrassenKnoten kn = (SINStrassenKnoten) knotenmap.get(knoten
						.getLocationCode());
				if (kn == null) {
					DEBUG
							.error("Zum ISS " + iss.getPid() + " kann kein Knoten gefunden werden."); //$NON-NLS-1$//$NON-NLS-2$
					continue;
				}
				kn.addSegment(ko);
			}
		}
	}

	/**
	 * Erzeugt die Stra&szlig;ensegmente.
	 */
	public void bildeStrassenSegmente() {
		int anzahlAss = 0;
		int anzahlIss = 0;

		// aeussere Strassensegmente erzeugen
		DEBUG.info("Die ueren Straensegmente werden erzeugt"); //$NON-NLS-1$
		anzahlAss = bildeAeussereStrassenSegmente();

		// innere Strassensegmente zu den aeusseren Strassensegmenten finden
		DEBUG
				.info("Die uere Straensegmente verbindenden inneren Straensegmente werden erzeugt"); //$NON-NLS-1$
		for (StrassenAbschnitt sa : assMenge) {

			ArrayList<Route> routen = importNetz
					.findeInnereStrassenSegmente(sa);

			for (LinkedList<StrassenAbschnitt> l : routen) {
				DEBUG.fine("Route:"); //$NON-NLS-1$
				for (StrassenAbschnitt sal : l) {
					DEBUG.fine(sal.toString());
					// sal.setZugeordnet();
				}
			}

			anzahlIss += bildeInnereStrassenSegmente(routen);
		}

		// innere Straensegmente fr die noch nicht zugeordneten Abschnitte
		// bilden
		DEBUG.info("Die fehlenden inneren Straensegmente werden erzeugt"); //$NON-NLS-1$
		for (StrassenAbschnitt abschnitt : strassenAbschnitte) {
			if (abschnitt.istISSAnfang(importNetz)) {
				ArrayList<Route> routen = importNetz
						.findeInnereStrassenSegmente(abschnitt);

				for (Route route : routen) {
					DEBUG.fine("Route:"); //$NON-NLS-1$
					for (StrassenAbschnitt sal : route) {
						DEBUG.fine(sal.toString());
					}
				}

				anzahlIss += bildeInnereStrassenSegmente(routen);
			}
		}

		DEBUG.info(anzahlAss + " uere Straensegmente erzeugt."); //$NON-NLS-1$
		DEBUG.info(anzahlIss + " innere Straensegmente erzeugt."); //$NON-NLS-1$

		segmentePidsKorrigieren();

	}

	/**
	 * Erzeugt die Stra&szlig;enteilsegmente.
	 */
	public void bildeStrassenTeilSegmente() {
		int anzahl = 0;

		DEBUG.info("Die Straenteilsegmente werden erzeugt"); //$NON-NLS-1$

		for (StrassenAbschnitt abschnitt : strassenAbschnitte) {
			if (!abschnitt.istZugeordnet()) {
				DEBUG.finest("bergehe nicht zugeordneten Abschnitt: "
						+ abschnitt);
				continue;
			}

			DEBUG.finest("bearbeite " + abschnitt);
			for (StrassenTeilAbschnitt sta : abschnitt.getTeilabschnitte()) {
				DEBUG.finest("bearbeite " + sta);
				String pid = null, name = null;

				// fr TEST-Zwecke
				String namenszusatz = ""; //$NON-NLS-1$
				if (SINProperties.getInstance().istAktiv(
						SINProperties.PROP_KEY_XMLDEBUG)) {
					namenszusatz += " (" + sta.strassenElement.getID() + ")"; //$NON-NLS-1$ //$NON-NLS-2$
				}

				pid = segmentFactory.bildePIDStrassenTeilSegment(sta);
				name = SegmentFactory.bildeSegmentNamen(pid, namenszusatz);

				SINSTS sts = new SINSTS(pid, name, sta);

				sts.setzteBetriebsKilometer(betriebsKilometer);
				sts.setRichtung(abschnitt.getRichtung());
				abschnitt.addSTS(sts);
				anzahl++;
			}
		}
		List<KonfigurationsObjekt> tmplist = new ArrayList<KonfigurationsObjekt>(
				konfigurationsObjekte.values());

		for (KonfigurationsObjekt ko : tmplist) {
			if ((ko instanceof SINASS) || (ko instanceof SINISS)) {
				addiereTeilsegmente((SINStrassenSegment) ko);
			} else {
				DEBUG.info("unklares Konfigurationsobjekt");
			}
		}
		DEBUG.info(anzahl + " Straenteilsegmente erzeugt."); //$NON-NLS-1$
	}

	/**
	 * &Uuml;berpr&uuml;ft die Locations auf G&uuml;ltigkeit.
	 * 
	 * @param von
	 *            VON-Location
	 * @param nach
	 *            NACHLocation
	 * @return true, wenn die beiden Locations g&uuml;ltig sind
	 * @throws ImportException
	 *             bei ung&uuml;ltigen Locations
	 * 
	 */
	private boolean checkLocations(final Location von, final Location nach)
			throws ImportException {
		if (von == Location.LOCATION_UNBEKANNT
				&& nach == Location.LOCATION_UNBEKANNT) {
			throw new ImportException(
					"Location-Problem: 2 'unbekannte' Locations"); //$NON-NLS-1$
		}

		if (von == null || nach == null) {
			throw new ImportException("Location-Problem: keine Location"); //$NON-NLS-1$
		}

		if (von.equals(nach)) {
			throw new ImportException(
					"Location-Problem: Anfangs- gleich Endlocation"); //$NON-NLS-1$
		}

		return true;
	}

	/**
	 * Exportiert alle erzeugten Konfigurationsobjekte
	 * {@link SINKonfigurationsBereich}.
	 * 
	 * @param kb
	 *            KonfigurationsBereich
	 */
	public void exportKB(SINKonfigurationsBereich kb) {
		for (KonfigurationsObjekt ko : konfigurationsObjekte.values()) {
			kb.addObjekt(ko);
		}
	}

	/**
	 * Gibt die erzeugten Konfigurationsobjekte zur&uuml;ck.
	 * 
	 * @return _objektmap
	 */
	public Map<String, KonfigurationsObjekt> getOjekteMap() {
		return konfigurationsObjekte;
	}

	/**
	 * Zentrale Methode zur Durchf&uuml;hrung des Importes und Erzeugung der
	 * internen Konfigurationsobjekte.
	 * 
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public void importiere() throws ImportException {
		bildeStrassenAbschnitte();
		apaktuell = new AufteilungspunkteAktuelleKonfiguration(
				aktuelleKonfiguration, dav);
		abschnitteAufteilen();
		bildeStrassenSegmente();
		segmenteTeilen();
		bildeStrassenTeilSegmente();
		bildeStrassen();
		bildeStrassenKnoten();
	}

	/**
	 * Korrigiert die PID und den Teil des Namens, der durch die PID bestimmt
	 * wird, dermassen, dass versucht wird, die kleinstm&ouml;gliche laufende
	 * Nummer als Postfix der PID zu nutzen.
	 * 
	 * @param ko
	 *            Konfigurationsobjekt
	 * @return korrigierte PID oder null, wenn nicht korrigiert wurde
	 */
	private String korrigierePid(KonfigurationsObjekt ko) {
		String pidstamm = ko.getPid()
				.substring(0, ko.getPid().lastIndexOf('.'));
		int laufendeNr = Integer.parseInt(ko.getPid().substring(
				ko.getPid().lastIndexOf('.') + 1));

		if (laufendeNr != 0) {
			String neuepid = pidstamm;

			int lfdnr = 0;
			do {
				neuepid = pidstamm + "." + lfdnr++;
				if (neuepid.equals(ko.getPid())) {
					return null;
				}
			} while (konfigurationsObjekte.containsKey(neuepid));

			return neuepid;
		}

		return null;
	}

	/**
	 * Die Methode korrigiert folgenden Effekt: <br>
	 * <p>
	 * Durch das sequenzielle Erzeugen der Stra&szlig;ensegmente entstehen bei
	 * &Auml;nderungen (ge&auml;nderter Polygonzug) Konfigurationsobjekte mit
	 * PIDs, die sich durch die laufende Nummer am Ende der PID vom Objekt der
	 * aktuellen Konfiguration unterscheiden.
	 * </p>
	 * <p>
	 * Erst nach dem alle Segmente erzeugt sind, kann entschieden werden, welche
	 * Objekte der aktuellen Konfiguration erhalten bleiben, bzw.
	 * &uuml;berschrieben werden sollen.
	 * </p>
	 * <p>
	 * Beispiel:
	 * </p>
	 * <p>
	 * In der aktuellen Konfiguration existiert das Stra&szlig;ensegment
	 * <code>ass.p.11111.p.22222.0</code>. Durch einen neuen Teilungspunkt
	 * (z.B. AQ) entsteht das gleiche Segment, aber mit ge&auml;ndertem
	 * Polygonzug. Das Segment der aktuellen Konfiguration wird deshalb nicht
	 * als identisch gefunden, das neue Segment erh&auml;lt die PID
	 * <code>ass.p.11111.p.22222.1</code>. Wenn in der neuen Konfiguration
	 * kein Objekt mit der PID <code>ass.p.11111.p.22222.0</code> existiert
	 * wird die PID des neuen Objektes in <code>ass.p.11111.p.22222.0</code>
	 * ge&auml;ndert.
	 * </p>
	 * <br>
	 * Dadurch wird sichergestellt, dass die Segmente immer mit einer laufenden
	 * Nummer '0' beginnen und die laufenden Nummern nicht mit jeder
	 * &Auml;nderung hochlaufen.
	 */
	private void segmentePidsKorrigieren() {
		List<KonfigurationsObjekt> tmplist = new ArrayList<KonfigurationsObjekt>(
				konfigurationsObjekte.values());

		for (KonfigurationsObjekt ko : tmplist) {
			if ((ko instanceof SINASS) || (ko instanceof SINISS)) {
				String korrigiertePid = korrigierePid(ko);
				if (korrigiertePid != null) {
					DEBUG.fine("Die PID '" + ko.getPid() + "' wird auf '"
							+ korrigiertePid + "' korrigiert");
					konfigurationsObjekte.remove(ko.getPid());
					((AbstraktStrassenSegment) ko)
							.korrigierePidUndNamen(korrigiertePid);
					addKonfigurationsObjekt(ko);
				}
			}
		}
	}

	/**
	 * Teilt die Segmente an den Aufteilungspunkten der aktuellen Konfiguration.
	 */
	private void segmenteTeilen() {
		List<AufteilungspunktAktuelleKonfiguration> punkte = apaktuell
				.getAufteilungspunkteLinienreferenz();

		List<AufteilungspunktAktuelleKonfiguration> nichtbenutzt = new ArrayList<AufteilungspunktAktuelleKonfiguration>();

		DEBUG
				.info("Die Straenabschnitte werden an den Aufteilungpunkten der aktuellen Konfiguration mit Linienreferenzen geteilt.");

		for (AufteilungspunktAktuelleKonfiguration punkt : punkte) {
			KonfigurationsObjekt ko = konfigurationsObjekte.get(punkt
					.getLinienReferenz());
			if (ko != null) {
				((AbstraktStrassenSegment) ko).aufteilenOffset(punkt
						.getLinienReferenzOffset());
			} else {
				nichtbenutzt.add(punkt);
			}
		}

		// nicht benutzte Punkte sichern
		aktuelleKonfiguration
				.setNichtVerarbeiteteAufteilungspunkteLinienreferenz(nichtbenutzt);
	}
}
