/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.geotools.feature.Feature;
import org.geotools.feature.FeatureType;

import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Polygon;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Punkt;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Stra&szlig;enteilabschnitt.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: StrassenTeilAbschnitt.java 21246 2010-01-22 14:39:03Z gieseler $
 * 
 */
public class StrassenTeilAbschnitt extends WGS84Polygon {
	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();


	/**
	 * Das zugeh&ouml;rige Stra&szlig;enelement.
	 */
	StrassenElement strassenElement;

	/**
	 * Die zugeordneten Eigenschaften-Features.
	 */
	List<Feature> eigenschaften = new ArrayList<Feature>();

	/**
	 * Die Liste der Streckenteilabschnitte.
	 */
	private final List<AsbStationierung> asbStationierungen = new ArrayList<AsbStationierung>();

	/**
	 * Erzeugt einen Stra&szlig;enteilabschnitt auf der Basis eines Polygonzuges
	 * f&uuml;r ein Stra&szlig;enelement mit bestimmten Eigenschaften.
	 * 
	 * @param polygon
	 *            der Polygonzug des Stra&szlig;enteilabschnittes
	 * @param strassenelement
	 *            das zugeh&ouml;rige Stra&szlig;enelement
	 * @param eigenschaften
	 *            Liste von diesen Stra&szlig;enteilabschnitt charakte-
	 *            risierende Eigenschaften
	 */
	public StrassenTeilAbschnitt(WGS84Polygon polygon,
			StrassenElement strassenelement, List<Feature> eigenschaften) {
		super(polygon.getKoordinaten());
		this.strassenElement = strassenelement;
		this.eigenschaften.addAll(eigenschaften);
	}

	/**
	 * Aktualisiert die Linien-Offsets der ASB-Stationierungsangaben.
	 */
	private void aktualisiereAsbOffsets() {
		double laengenkorrektur = berechneLaengenkorrektur();
		int linienoffset = 0;

		for (int i = 0; i < asbStationierungen.size(); i++) {
			AsbStationierung asb = asbStationierungen.get(i);

			if(asb.getVerkehrsRichtung() == Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG) {
				if (i == 0) {
					asb.offset = 0;
				} else {
					asb.offset = (int) (linienoffset * laengenkorrektur);
				}

				linienoffset += asb.getStationierungEnde()
					- asb.getStationierungAnfang();
			}
			else if (asb.getVerkehrsRichtung() == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
				if (i == asbStationierungen.size()-1) {
					asb.offset = 0;
				} else {
					linienoffset += asb.getStationierungEnde()
					- asb.getStationierungAnfang();
					asb.offset = (int) (laengeExakt() - (linienoffset * laengenkorrektur));
				}
			}
			else {
				DEBUG.error("ASB-Stationierungsrichtung unklar");
			}
		}

	}

	/**
	 * Aktualisiert die Eigenschaften.
	 * 
	 * @param punkt
	 *            Teilungspunkt
	 */
	public void aktualisiereFeatures(Teilungspunkt punkt) {
		Map<FeatureType, Feature> neu = new HashMap<FeatureType, Feature>();

		// entferne nicht mehr gltige
		for (Feature fb : eigenschaften) {
			boolean keep = true;
			for (Feature fe : punkt.getFeaturesEnde()) {
				if (fb.getFeatureType().equals(fe.getFeatureType())) {
					keep = false;
					break;
				}
			}

			if (keep) {
				neu.put(fb.getFeatureType(), fb);
			}
		}

		// neue berschreiben
		for (Feature fe : punkt.getFeaturesBeginn()) {
			neu.put(fe.getFeatureType(), fe);
		}

		eigenschaften = new ArrayList<Feature>(neu.values());
	}

	/**
	 * Berechnet den L&auml;ngenkorrekturfaktor f&uuml;r den Abschnitt als
	 * Verh&auml;ltnis von Polygonl&auml;nge zu ASB-Stationierungsl&auml;nge.
	 * 
	 * @return L&auml;ngenkorrekturfaktor als Verh&auml;ltnis von
	 *         Polygonl&auml;nge zu ASB-Stationierungsl&auml;nge.
	 */
	private double berechneLaengenkorrektur() {
		// berechne gesamte ASB-Lnge
		int asblaenge = 0;
		for (AsbStationierung asb : asbStationierungen) {
			asblaenge += asb.getStationierungEnde()
					- asb.getStationierungAnfang();
		}

		if (asblaenge != 0) {
			return laenge() / asblaenge;
		}

		return 1.0;
	}

	/**
	 * Gibt die Liste der ASB-Stationierungen zur&uuml;ck.
	 * 
	 * @return Liste von {@link AsbStationierung}
	 */
	public List<AsbStationierung> getAsbStationierungen() {
		return asbStationierungen;
	}

	/**
	 * Gibt die Liste der den Stra&szlig;enteilabschnitt charakterisierenden
	 * Eigenschaften zur&uuml;ck.
	 * 
	 * @return Eigenschaften als Liste von {@link Feature}
	 */
	public List<Feature> getEigenschaften() {
		return eigenschaften;
	}

	/**
	 * Gibt das zugeh&ouml;rige Stra&szlig;enelement zur&uuml;ck.
	 * 
	 * @return {@link StrassenElement}
	 */
	public StrassenElement getStrassenElement() {
		return strassenElement;
	}

	/**
	 * Setzt eine ASB-Stationierung.
	 * 
	 * @param stationierung
	 *            AsbStationierung
	 */
	public void setAsbStationierung(AsbStationierung stationierung) {
		asbStationierungen.add(stationierung);
		aktualisiereAsbOffsets();
	}

	/**
	 * Setzt die ASB-Stationierungen.
	 * 
	 * @param stationierungen
	 *            Liste von {@link AsbStationierung}
	 */
	public void setAsbStationierungen(List<AsbStationierung> stationierungen) {
		asbStationierungen.addAll(stationierungen);
		aktualisiereAsbOffsets();
	}

	/**
	 * Teilt den Teilabschnitt an einem Punkt.
	 * 
	 * @param punkt
	 *            Punkt
	 * @return Liste von neu entstandenen Teilabschnitten
	 */
	public List<StrassenTeilAbschnitt> teilen(WGS84Punkt punkt) {
		double laengenkorrektur = berechneLaengenkorrektur();

		// Polygone berechnen
		WGS84Punkt teilungspunkt = this.bildPunkt(punkt);
		// double teilungspunktoffset = this.berecheneOffset(teilungspunkt);

		WGS84Polygon org = new WGS84Polygon(this.getKoordinaten());
		WGS84Polygon anfang = org.anfangAbschneiden(teilungspunkt);

		List<StrassenTeilAbschnitt> neueAbschnitte = new ArrayList<StrassenTeilAbschnitt>();

		StrassenTeilAbschnitt neuAnfang = new StrassenTeilAbschnitt(anfang,
				strassenElement, eigenschaften);
		StrassenTeilAbschnitt neuEnde = new StrassenTeilAbschnitt(org,
				strassenElement, eigenschaften);

		// ASB-Stationierungen berechnen
		boolean asbAnfangBeendet = false;
		int asblaenge = 0;
		for (AsbStationierung asb : this.asbStationierungen) {
			asblaenge += asb.getStationierungEnde()
					- asb.getStationierungAnfang();
			if (!asbAnfangBeendet) {
				if ((asblaenge * laengenkorrektur) > neuAnfang.laenge()) {
					asbAnfangBeendet = true;
					double restlaenge = neuAnfang.laenge() - asb.getOffset();
					int asboffsetpunkt = ((int) (restlaenge / laengenkorrektur))
							+ asb.getStationierungAnfang();
					AsbStationierung korr = asb.clone();
					asb.stationierungEnde = korr.stationierungAnfang = asboffsetpunkt;

					neuAnfang.setAsbStationierung(asb);
					neuEnde.setAsbStationierung(korr);
				} else {
					neuAnfang.setAsbStationierung(asb);
				}
			} else {
				neuEnde.setAsbStationierung(asb);
			}

		}
		neueAbschnitte.add(neuAnfang);
		neueAbschnitte.add(neuEnde);

		return neueAbschnitte;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.util.AbstractCollection#toString()
	 */
	@Override
	public String toString() {
		String ret = null;

		ret = "StrassenTeilabschnitt[" + strassenElement.getID() + "]"; //$NON-NLS-1$

		return ret;
	}
}
