/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig.okstra;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.TreeMap;

import javax.xml.parsers.ParserConfigurationException;

import org.geotools.data.DataStore;
import org.geotools.data.DefaultQuery;
import org.geotools.data.FeatureReader;
import org.geotools.data.FeatureSource;
import org.geotools.data.FeatureWriter;
import org.geotools.data.LockingManager;
import org.geotools.data.Query;
import org.geotools.data.Transaction;
import org.geotools.data.wfs.BWSIN_WFSDataStore;
import org.geotools.data.wfs.BWSIN_WFSDataStoreFactory;
import org.geotools.data.wfs.BWSIN_WFSFileDataStore;
import org.geotools.data.wfs.WFSDataStoreFactory;
import org.geotools.factory.FactoryRegistryException;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureType;
import org.geotools.feature.IllegalAttributeException;
import org.geotools.feature.SchemaException;
import org.geotools.filter.Filter;
import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.opengis.referencing.operation.TransformException;
import org.opengis.spatialschema.geometry.MismatchedDimensionException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;

import de.bsvrz.kex.bwsin.konfig.ImportException;
import de.bsvrz.kex.bwsin.konfig.okstra.parser.SINFeatureHandler;
import de.bsvrz.kex.bwsin.konfig.okstra.parser.SINXMLHandler;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Polygon;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modul OKSTRA.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: Okstra.java 10427 2008-07-10 12:09:37Z gieseler $
 * 
 */
public class Okstra implements DataStore {

	/**
	 * Instanz der Klasse.
	 */
	private static Okstra instance = null;

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * Findet den Wert eines Attributes.
	 * 
	 * @param node
	 *            Knoten
	 * @param attr
	 *            Attributname
	 * @return Wert des Attributes oder null
	 */
	private static String findAttrElementValue(Node node, String attr) {
		if (node.getNodeName().equals(attr)) {
			NodeList childs = node.getChildNodes();
			for (int i = 0; i < childs.getLength(); i++) {
				if (childs.item(i).getNodeType() == Node.TEXT_NODE) {
					return childs.item(i).getNodeValue();
				}
			}
			if (node.hasAttributes()) {
				return ((Element) node).getAttribute("xlink:href");
			}
		}

		NodeList childs = node.getChildNodes();
		for (int i = 0; i < childs.getLength(); i++) {
			if (childs.item(i).getNodeType() == Node.ELEMENT_NODE) {
				String ret = findAttrElementValue(childs.item(i), attr);
				if (ret != null) {
					return ret;
				}
			}
		}

		return null;
	}

	/**
	 * Findet alle Attributwerte zu einem Attribut.
	 * 
	 * @param node
	 *            Knoten
	 * @param attr
	 *            Attributname
	 * @return Liste von Attributwerten
	 */
	private static List<String> findAttrElementValues(Node node, String attr) {
		List<String> ret = new ArrayList<String>();
		String[] attrs = attr.split("\\.");

		if (node.getNodeName().equals(attrs[0])) {
			if (attrs.length > 1) {
				ret.addAll(findAttrElementValues(node, attrs[1]));
			} else {
				NodeList childs = node.getChildNodes();
				for (int i = 0; i < childs.getLength(); i++) {
					if (childs.item(i).getNodeType() == Node.TEXT_NODE) {
						ret.add(childs.item(i).getNodeValue());
						return ret;
					}
				}
				if (node.hasAttributes()) {
					ret.add(((Element) node).getAttribute("xlink:href"));
				}

				return ret;
			}
		}

		NodeList childs = node.getChildNodes();
		for (int i = 0; i < childs.getLength(); i++) {
			if (childs.item(i).getNodeType() == Node.ELEMENT_NODE) {
				List<String> f = findAttrElementValues(childs.item(i), attr);
				if (f != null) {
					ret.addAll(f);
				}
			}
		}

		return ret;
	}

	/**
	 * Gibt die Koordinatenliste eines Features als WGS84-Polygon zur&uuml;ck.
	 * 
	 * @param f
	 *            {@link Feature}
	 * @return {@link WGS84Polygon}
	 */
	public static WGS84Polygon getFeatureKoordinatenAlsWGS84Polygon(Feature f) {
		Coordinate[] c = getFeatureKoordinatenWGS84(f);
		double[] x = new double[c.length];
		double[] y = new double[c.length];

		for (int i = 0; i < c.length; i++) {
			x[i] = c[i].x;
			y[i] = c[i].y;
		}

		return new WGS84Polygon(x, y);
	}

	/**
	 * Gibt die Koordinatenliste eines Features als WGS84-Koordninaten
	 * zur&uuml;ck.
	 * 
	 * @param f
	 *            {@link Feature}
	 * @return Array von Koordinaten
	 */
	public static Coordinate[] getFeatureKoordinatenWGS84(Feature f) {
		FeatureType ft = f.getFeatureType();
		// GeometryFactory gf = f.getDefaultGeometry().getFactory();
		CoordinateReferenceSystem sourceCrs = ft.getDefaultGeometry()
				.getCoordinateSystem();
		CoordinateReferenceSystem destCrs = DefaultGeographicCRS.WGS84;

		if (sourceCrs == null) { // MIF!
			try {
				sourceCrs = CRS.decode("EPSG:32632"); // WGS84/UTM zone 32N
				// //$NON-NLS-1$
			} catch (FactoryRegistryException e1) {
				e1.printStackTrace();
			} catch (FactoryException e1) {
				e1.printStackTrace();
			}
		}

		if (sourceCrs != null && !sourceCrs.equals(destCrs)) {
			MathTransform trans;

			try {
				trans = CRS.findMathTransform(sourceCrs,
						DefaultGeographicCRS.WGS84, true);
				Geometry newGeom = JTS.transform(f.getDefaultGeometry(), trans);
				return newGeom.getCoordinates();
			} catch (FactoryException e) {
				e.printStackTrace();
			} catch (MismatchedDimensionException e) {
				e.printStackTrace();
			} catch (TransformException e) {
				e.printStackTrace();
			}
		} else {
			return f.getDefaultGeometry().getCoordinates();
		}

		return null;
	}

	/**
	 * Extrahiert die Feature-Referenz aus dem XLINK-Attribut und gibt das
	 * referenzierte Feature zur&uuml;ck.
	 * 
	 * @param attribute
	 *            ein eine Feature-Referenz enthaltendes Feature-Attribut
	 * @return {@link Feature}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public static Feature getFeatureReferenz(Object attribute)
			throws ImportException {
		// urn:x-vrzbw:Strecke.0900b8a4a1
		if (attribute == null) {
			throw new IllegalArgumentException(
					"'getFeatureReferenz' Attribute darf nicht null sein"); //$NON-NLS-1$
		}
		String ftype = getRefFeatureType((String) attribute);
		String fid = getRefFeatureId((String) attribute);

		return Okstra.getInstance().getFeature(ftype, fid);
	}

	/**
	 * Gibt den Wert eines Attributes eines Features zur&uuml;ck.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param id
	 *            Feature-ID
	 * @param attr
	 *            Name des Attributes
	 * 
	 * @return Wert des Attributes
	 */
	private static String getIdValue(OkstraFeature of, String id, String attr) {
		String ename = of.getPrefix() + id.substring(0, id.indexOf('.'));
		NodeList nl = of.getDocument().getElementsByTagName(ename);
		Node found = null;

		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			NamedNodeMap attrs = n.getAttributes();

			Node fid = attrs.getNamedItem("fid");
			if (fid.getNodeValue().equals(id)) {
				found = n;
				break;
			}
		}

		if (found != null) {
			return findAttrElementValue(found, attr);
		}

		return null;
	}

	/**
	 * Gibt alle Werte eines Attributes eines Features zur&uuml;ck.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param id
	 *            Feature-ID
	 * @param attr
	 *            Name des Attributes
	 * 
	 * @return Werte des Attributes
	 */
	private static Object[] getIdValues(OkstraFeature of, String id, String attr) {

		Node found = of.getNode(id);

		if (found != null) {
			return findAttrElementValues(found, attr).toArray();
		}

		return null;
	}

	/**
	 * Gibt alle Werte eines Attributes eines Features zur&uuml;ck.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param id
	 *            Feature-ID
	 * @param attname
	 *            Name des Attributes
	 * 
	 * @return Werte des Attributes
	 */
	private static List<Object> getIdValues2(OkstraFeature of, String id,
			String attname) {

		SINFeatureHandler sfh = of.getFeatures().get(id);

		return sfh.getAttributeWerte(attname);
	}

	/**
	 * Gibt die Instanz auf das OKSTRA-Objekt zur&uuml;ck.
	 * 
	 * @return Instanz
	 */
	public static Okstra getInstance() {
		return instance;
	}

	/**
	 * Gibt ein ganzzahliges Attribut eines Features zur&uuml;ck.
	 * 
	 * @param f
	 *            {@link Feature}
	 * @param name
	 *            Name des Attributes
	 * @return Wert
	 * @throws IllegalAttributeConversionException
	 *             wenn der Wert des Attributes nicht in eine Ganzzahl
	 *             konvertiert werden kann
	 */
	public static int getIntAttribute(Feature f, String name)
			throws IllegalAttributeConversionException {
		Object obj = f.getAttribute(name);

		if (obj == null) {
			throw new IllegalArgumentException("Attribute '" + name + //$NON-NLS-1$
					"' existiert nicht im Featuretyp '" + //$NON-NLS-1$
					f.getFeatureType().getTypeName() + "'"); //$NON-NLS-1$
		}

		try {
			if (obj instanceof Number) {
				Number number = (Number) obj;

				return (number.intValue());
			} else if (obj instanceof String) {
				NumberFormat nf = NumberFormat.getInstance();
				Number number = nf.parse((String) obj);

				return (int) Math.round(number.doubleValue());
			} else if (obj instanceof Double) {
				Number number = (Number) obj;

				return (int) Math.round(number.doubleValue());
			}

		} catch (ParseException e) {
			/**
			 * 
			 */
		}
		throw new IllegalAttributeConversionException("Attribute '" + name + //$NON-NLS-1$
				"' kann nicht in integer Konvertiert werden"); //$NON-NLS-1$
	}

	/**
	 * Extrahiert die Feature-ID aus dem XPATH-Attribut und gibt diese als
	 * String zur&uuml;ck.
	 * 
	 * @param xpath
	 *            ein eine Feature-Referenz enthaltendes Feature-Attribut
	 * @return der referenzierte Feature-Typ
	 */
	public static String getRefFeatureId(String xpath) {
		int index = xpath.lastIndexOf(":"); //$NON-NLS-1$
		return xpath.substring(index + 1);
	}

	/**
	 * Extrahiert den Feature-Typ aus dem XPATH-Attribut und gibt diesen als
	 * String zur&uuml;ck.
	 * 
	 * @param xpath
	 *            ein eine Feature-Referenz enthaltendes Feature-Attribut
	 * @return der referenzierte Feature-Typ
	 */
	public static String getRefFeatureType(String xpath) {
		int index1 = xpath.lastIndexOf(":"); //$NON-NLS-1$
		int index2 = xpath.lastIndexOf("."); //$NON-NLS-1$
		return "okstra:" + xpath.substring(index1 + 1, index2); //$NON-NLS-1$
	}

	/**
	 * Datastore.
	 */
	DataStore dataStore;

	/**
	 * Feature-Cache.
	 */
	HashMap<String, Map<String, Feature>> cache = new HashMap<String, Map<String, Feature>>();

	/**
	 * Erzeugt eine Instanz zum Zugriff auf die OKSTRA-Datenbasis.
	 * 
	 * @param prop
	 *            Properties
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public Okstra(final Properties prop) throws ImportException {
		final BWSIN_WFSDataStoreFactory mdf = new BWSIN_WFSDataStoreFactory();
		try {
			dataStore = mdf.createDataStore(prepareWfsParam(prop));
		} catch (final IOException e) {
			throw new ImportException(
					"Die WFS-Datenquelle kann nicht initialisiert werden ("
							+ e.getClass().getName() + ": " + e.getMessage()
							+ ")");
		}

		instance = this;
	}

	/**
	 * Erzeugt eine Hash-Tabelle der Features.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param id
	 *            Feature-ID
	 * @param doc
	 *            Dukument
	 */
	private void buildHash(OkstraFeature of, String id, Document doc) {

		String ename = of.getPrefix() + id.substring(0, id.indexOf('.'));
		NodeList nl = of.getDocument().getElementsByTagName(ename);

		int nllength = nl.getLength(), read = 0;

		Map<String, Node> nodemap = new HashMap<String, Node>(nllength);

		for (int i = 0; i < nllength; i++) {
			read++;
			if ((read % 1000) == 0) {
				DEBUG.finest("gelesen: " + read + " von " + nllength
						+ " Datenstze(n)");
			}

			Node n = nl.item(i);
			NamedNodeMap attrs = n.getAttributes();

			Node fid = attrs.getNamedItem("fid");
			nodemap.put(fid.getNodeValue(), n.cloneNode(true));
		}

		of.setMap(nodemap);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#createSchema(org.geotools.feature.FeatureType)
	 */
	public void createSchema(FeatureType arg0) throws IOException {
		dataStore.createSchema(arg0);
	}

	/**
	 * Findet das 'echte' Feature zu einer Abschnitt-oder-Ast Referenz.
	 * 
	 * @param iD
	 *            Abschnitt-oder-Ast
	 * @return Abschnitt- bzw. Ast-Feature
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public Feature findAoA(String iD) throws ImportException {
		Feature f = getFeature(OkstraFeatures.ABSCHNITT, iD);
		if (f == null) {
			return getFeature(OkstraFeatures.AST, iD);
		}

		return f;
	}

	/**
	 * Gibt ein Attribut eines Features als double-Wert zur&uuml;ck.
	 * 
	 * @param f
	 *            {@link Feature}
	 * @param name
	 *            Name des Attributes
	 * @return Wert
	 * @throws IllegalAttributeConversionException
	 *             wenn der Wert des Attributes nicht in double konvertiert
	 *             werden kann
	 */
	public double getDoubleAttribute(Feature f, String name)
			throws IllegalAttributeConversionException {
		Object obj = f.getAttribute(name);

		if (obj == null) {
			throw new IllegalArgumentException("Attribute '" + name + //$NON-NLS-1$
					"' existiert nicht im Featuretyp '" + //$NON-NLS-1$
					f.getFeatureType().getTypeName() + "'"); //$NON-NLS-1$
		}

		if (obj instanceof Number) {
			Number number = (Number) obj;

			return (number.doubleValue());
		}
		throw new IllegalAttributeConversionException("Attribute '" + name + //$NON-NLS-1$
				"' kann nicht in integer Konvertiert werden"); //$NON-NLS-1$
	}

	/**
	 * Gibt ein Feature eines abgefragten Types zur&uuml;ck.
	 * 
	 * @param ofeature
	 *            das abzufragende Feature
	 * @param fid
	 *            Feature-ID
	 * @return {@link Feature}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public Feature getFeature(OkstraFeature ofeature, String fid)
			throws ImportException {
		Map<String, Feature> smap;
		if (cache.containsKey(ofeature.getFeatureOkstraName())) {
			smap = cache.get(ofeature.getFeatureOkstraName());
		} else {
			smap = getFeatureMap(ofeature.getFeatureOkstraName());
		}

		return smap.get(fid);
	}

	/**
	 * Gibt ein Features eines abgefragten Types zur&uuml;ck.
	 * 
	 * @param ftype
	 *            der Name des Feature-Typs
	 * @param fid
	 *            Feature-ID
	 * @return {@link Feature}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public Feature getFeature(String ftype, String fid) throws ImportException {
		Map<String, Feature> smap;
		if (cache.containsKey(ftype)) {
			smap = cache.get(ftype);
		} else {
			smap = getFeatureMap(ftype);
		}

		return smap.get(fid);
	}

	/**
	 * Bildet eine Map von Features eines abgefragten Typs.
	 * 
	 * @param featuretype
	 *            der abgefragte Feature-Typ
	 * @return Feature-Tabelle
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public Map<String, Feature> getFeatureMap(String featuretype)
			throws ImportException {
		if (cache.containsKey(featuretype)) {
			return (cache.get(featuretype));
		}

		DEBUG.info("Feature '" + featuretype + "' wird vom WFS gelesen"); //$NON-NLS-1$//$NON-NLS-2$

		Map<String, Feature> ret = new TreeMap<String, Feature>();

		Query query = new DefaultQuery(featuretype);
		FeatureReader reader;
		int read = 0;
		try {
			reader = dataStore.getFeatureReader(query, Transaction.AUTO_COMMIT);
			while (reader.hasNext()) {
				Feature f = reader.next();
				read++;
				if ((read % 100) == 0) {
					DEBUG.finest("gelesen: " + read + " Datenstze"); //$NON-NLS-1$//$NON-NLS-2$
				}

				ret.put(f.getID(), f);
			}
			cache.put(featuretype, ret);
		} catch (IOException e) {
			throw new ImportException(e.getMessage());
		} catch (NoSuchElementException e) {
			throw new ImportException(e.getMessage());
		} catch (IllegalAttributeException e) {
			throw new ImportException(e.getMessage());
		}

		DEBUG.info("Feature '" + featuretype + "' wurde vom WFS gelesen"); //$NON-NLS-1$ //$NON-NLS-2$
		return ret;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getFeatureReader(org.geotools.data.Query,
	 *      org.geotools.data.Transaction)
	 */
	public FeatureReader getFeatureReader(Query arg0, Transaction arg1)
			throws IOException {
		return dataStore.getFeatureReader(arg0, arg1);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getFeatureSource(java.lang.String)
	 */
	public FeatureSource getFeatureSource(String arg0) throws IOException {
		return dataStore.getFeatureSource(arg0);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getFeatureWriter(java.lang.String,
	 *      org.geotools.filter.Filter, org.geotools.data.Transaction)
	 */
	public FeatureWriter getFeatureWriter(String arg0, Filter arg1,
			Transaction arg2) throws IOException {
		throw new UnsupportedOperationException();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getFeatureWriter(java.lang.String,
	 *      org.geotools.data.Transaction)
	 */
	public FeatureWriter getFeatureWriter(String arg0, Transaction arg1)
			throws IOException {
		throw new UnsupportedOperationException();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getFeatureWriterAppend(java.lang.String,
	 *      org.geotools.data.Transaction)
	 */
	public FeatureWriter getFeatureWriterAppend(String arg0, Transaction arg1)
			throws IOException {
		throw new UnsupportedOperationException();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getLockingManager()
	 */
	public LockingManager getLockingManager() {
		return dataStore.getLockingManager();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getSchema(java.lang.String)
	 */
	public FeatureType getSchema(String arg0) throws IOException {
		return dataStore.getSchema(arg0);
	}

	/**
	 * Spezialfunktion zum Lesen von Attributen, die von GeoTools nicht
	 * bereitgestellt werden k&ouml;nnen.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param fid
	 *            Feature-ID
	 * @param attname
	 *            Attributname
	 * @return Attributwert
	 */
	public String getSpecialAttribute(final OkstraFeature of, final String fid,
			final String attname) {

		if (of.getDocument() == null) {
			Query query = new DefaultQuery(of.getFeatureOkstraName());
			try {
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " wird gelesen");
				Document doc = null;
				if (dataStore instanceof BWSIN_WFSDataStore) {
					doc = ((BWSIN_WFSDataStore) dataStore).getDocument(query);
				} else if (dataStore instanceof BWSIN_WFSFileDataStore) {
					doc = ((BWSIN_WFSFileDataStore) dataStore)
							.getDocument(query);
				} else {
					throw new IllegalArgumentException(
							"DataStore untersttzt die Operation nicht!");
				}

				of.setDocument(doc);
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " wurde gelesen");
				buildHash(of, fid, doc);
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " HashMap erzeugt");
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (SAXException e) {
				e.printStackTrace();
			} catch (ParserConfigurationException e) {
				e.printStackTrace();
			}
		}

		return getIdValue(of, fid, attname);
	}

	/**
	 * Spezialfunktion zum Lesen von Attributen, die von GeoTools nicht
	 * bereitgestellt werden k&ouml;nnen.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param fid
	 *            Feature-ID
	 * @param attname
	 *            Attributname
	 * @return Attributwert
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public String getSpecialAttribute2(OkstraFeature of, String fid,
			String attname) throws ImportException {
		Object[] obj = getSpecialAttributeList2(of, fid, attname);

		if (obj.length > 1) {
			throw new ImportException("mehrere Attribute verfgbar");
		}

		if (obj.length == 0) {
			return null;
		}

		return (String) obj[0];
	}

	/**
	 * Spezialfunktion zum Lesen von mehreren Attributen zum gleichen Namen, die
	 * von GeoTools nicht bereitgestellt werden k&ouml;nnen.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param fid
	 *            Feature-ID
	 * @param attname
	 *            Attributname
	 * @return Attributwert
	 */
	public Object[] getSpecialAttributeList(OkstraFeature of, String fid,
			String attname) {

		if (of.getDocument() == null) {
			Query query = new DefaultQuery(of.getFeatureOkstraName());
			try {
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " wird gelesen");
				Document doc = null;
				if (dataStore instanceof BWSIN_WFSDataStore) {
					doc = ((BWSIN_WFSDataStore) dataStore).getDocument(query);
				} else if (dataStore instanceof BWSIN_WFSFileDataStore) {
					doc = ((BWSIN_WFSFileDataStore) dataStore)
							.getDocument(query);
				} else {
					throw new IllegalArgumentException(
							"DataStore untersttzt die Operation nicht");
				}

				of.setDocument(doc);
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " wurde gelesen");
				buildHash(of, fid, doc);
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " HashMap erzeugt");
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (SAXException e) {
				e.printStackTrace();
			} catch (ParserConfigurationException e) {
				e.printStackTrace();
			}
		}

		return getIdValues(of, fid, attname);
	}

	/**
	 * Spezialfunktion zum Lesen von mehreren Attributen zum gleichen Namen, die
	 * von GeoTools nicht bereitgestellt werden k&ouml;nnen.
	 * 
	 * @param of
	 *            OKSTRA-Feature
	 * @param fid
	 *            Feature-ID
	 * @param attname
	 *            Attributname
	 * @return Attributwert
	 */
	public Object[] getSpecialAttributeList2(OkstraFeature of, String fid,
			String attname) {

		if (of.getFeatures().size() == 0) {
			Query query = new DefaultQuery(of.getFeatureOkstraName());
			try {
				DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
						+ " wird gelesen");
				InputSource source = null;

				if (dataStore instanceof BWSIN_WFSDataStore) {
					source = ((BWSIN_WFSDataStore) dataStore).getSource(query);
				} else if (dataStore instanceof BWSIN_WFSFileDataStore) {
					source = ((BWSIN_WFSFileDataStore) dataStore)
							.getSource(query);
				} else {
					throw new IllegalArgumentException(
							"DataStore untersttzt die Operation nicht");
				}

				readFeatures(of, source);
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (SAXException e) {
				e.printStackTrace();
			} catch (ParserConfigurationException e) {
				e.printStackTrace();
			}
		}

		return ((Collection<?>) getIdValues2(of, fid, attname)).toArray();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getTypeNames()
	 */
	public String[] getTypeNames() throws IOException {
		return dataStore.getTypeNames();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#getView(org.geotools.data.Query)
	 */
	public FeatureSource getView(Query arg0) throws IOException,
			SchemaException {
		return dataStore.getView(arg0);
	}

	/**
	 * Erzeugt die Parameter-Map f&uuml;r den WFS-Datastore aus den
	 * &uuml;bergebenen Properties.
	 * 
	 * @param prop
	 *            Properties
	 * @return Parameter-Tabelle
	 * @throws IOException
	 *             bei Ausnahmen
	 */
	@SuppressWarnings("nls")
	private Map<String, Object> prepareWfsParam(Properties prop)
			throws IOException {
		Map<String, Object> params = new HashMap<String, Object>();

		if (!prop.containsKey("SERVER_URL")) {
			throw new IOException("'SERVER_URL' Parameter erwartet");
		}

		DEBUG.info("OKSTRA-Datastore: " + prop.get("SERVER_URL"));

		params.put(WFSDataStoreFactory.URL.key, prop.get("SERVER_URL"));

		if (prop.containsKey("USE_GET")) {
			prop.getProperty("USE_GET", "");
		}

		params.put(WFSDataStoreFactory.PROTOCOL.key, Boolean.FALSE); // false
		// ==
		// GET,
		// true
		// ==
		// POST

		params.put(WFSDataStoreFactory.TIMEOUT.key, Integer.parseInt(prop
				.getProperty("TIMEOUT", "100000")));

		params.put(WFSDataStoreFactory.BUFFER_SIZE.key, Integer.parseInt(prop
				.getProperty("BUFFER_SIZE", "100000")));

		if (prop.containsKey("FILESTORE")) {
			params.put(BWSIN_WFSDataStoreFactory.FILETYPE.key, prop
					.get("FILESTORE"));
		}

		return params;
	}

	/**
	 * Liest Features von einer Quelle.
	 * 
	 * @param of
	 *            OKSTR-Feature
	 * @param source
	 *            Quelle
	 */
	private void readFeatures(final OkstraFeature of, final InputSource source) {
		DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
				+ " wurde gelesen");
		XMLReader xr;
		try {
			xr = XMLReaderFactory.createXMLReader();

			SINXMLHandler handler = new SINXMLHandler();
			xr.setContentHandler(handler);
			xr.setErrorHandler(handler);

			xr.parse(source);
			of.setFeatures(handler.getFeatures());
			DEBUG.info("Feature(Special): " + of.getFeatureOkstraName()
					+ " HashMap erzeugt");
		} catch (SAXException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see org.geotools.data.DataStore#updateSchema(java.lang.String,
	 *      org.geotools.feature.FeatureType)
	 */
	public void updateSchema(String arg0, FeatureType arg1) throws IOException {
		dataStore.updateSchema(arg0, arg1);
	}

}
